/*
 * Copyright (c) 1999-2000 Caucho Technology.  All rights reserved.
 *
 * Caucho Technology permits redistribution, modification and use
 * of this file in source and binary form ("the Software") under the
 * Caucho Developer Source License ("the License").  In particular, the following
 * conditions must be met:
 *
 * 1. Each copy or derived work of the Software must preserve the copyright
 *    notice and this notice unmodified.
 *
 * 2. Redistributions of the Software in source or binary form must include 
 *    an unmodified copy of the License, normally in a plain ASCII text
 *
 * 3. The names "Resin" or "Caucho" are trademarks of Caucho Technology and
 *    may not be used to endorse products derived from this software.
 *    "Resin" or "Caucho" may not appear in the names of products derived
 *    from this software.
 *
 * 4. Caucho Technology requests that attribution be given to Resin
 *    in any manner possible.  We suggest using the "Resin Powered"
 *    button or creating a "powered by Resin(tm)" link to
 *    http://www.caucho.com for each page served by Resin.
 *
 * This Software is provided "AS IS," without a warranty of any kind. 
 * ALL EXPRESS OR IMPLIED REPRESENTATIONS AND WARRANTIES, INCLUDING ANY
 * IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE
 * OR NON-INFRINGEMENT, ARE HEREBY EXCLUDED.

 * CAUCHO TECHNOLOGY AND ITS LICENSORS SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE OR ANY THIRD PARTY AS A RESULT OF USING OR
 * DISTRIBUTING SOFTWARE. IN NO EVENT WILL Caucho OR ITS LICENSORS BE LIABLE
 * FOR ANY LOST REVENUE, PROFIT OR DATA, OR FOR DIRECT, INDIRECT, SPECIAL,
 * CONSEQUENTIAL, INCIDENTAL OR PUNITIVE DAMAGES, HOWEVER CAUSED AND
 * REGARDLESS OF THE THEORY OF LIABILITY, ARISING OUT OF THE USE OF OR
 * INABILITY TO USE SOFTWARE, EVEN IF HE HAS BEEN ADVISED OF THE POSSIBILITY
 * OF SUCH DAMAGES.      
 */

/*
 * config.c is responsible for scanning the parsed registry and grabbing
 * relevant data.
 *
 * Important data include the web-app and the servlet-mapping so any filter
 * can properly dispatch the request.
 *
 * Also, config.c needs to grab the srun and srun-backup blocks to properly
 * send the requests to the proper JVM.
 */

#ifdef WIN32
#include <winsock2.h>
#else
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <dirent.h>
#endif
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <sys/stat.h>
#include "cse.h"

#define CACHE_SIZE 16384

typedef struct hash_t {
  char *host;
  int port;
  char *uri;
  int is_match;
  int count;
} hash_t;

static hash_t g_url_cache[CACHE_SIZE];

static void
cse_add_unique_location(config_t *config, char *host, int port, char *prefix,
                        char *suffix, int is_exact)
{
  location_t *loc;

  for (loc = config->locations; loc; loc = loc->next) {
    if ((host == 0) != (loc->host == 0))
      continue;
    else if (host && strcmp(host, loc->host))
      continue;
    else if (port != loc->port)
      continue;
    else if (is_exact != loc->is_exact)
      continue;
    else if ((prefix == 0) != (loc->prefix == 0))
      continue;
    else if (prefix && strcmp(prefix, loc->prefix))
      continue;
    else if ((suffix == 0) != (loc->suffix == 0))
      continue;
    else if (suffix && strcmp(suffix, loc->suffix))
      continue;

    return;
  }

  loc = (location_t *) cse_alloc(config->p, sizeof(location_t));

  memset(loc, 0, sizeof(location_t));

  loc->next = config->locations;
  config->locations = loc;

  loc->host = host;
  loc->port = port;
  loc->prefix = prefix;
  loc->suffix = suffix;
  loc->is_exact = is_exact;

  LOG(("loc %s:%d %s %s %x\n", 
       loc->host ? loc->host : "(null)",
       loc->port,
       loc->prefix ? loc->prefix : "(null)",
       loc->suffix ? loc->suffix : "(null)",
       loc->next));
}

/**
 * Add a location pattern to the list of recognized locations
 *
 * @param config the configuration
 * @param host the host for the pattern
 * @param prefix the web-app prefix
 * @param pattern the url-pattern to match
 *
 * @return the new location
 */
static void
cse_add_location(config_t *config, char *host, char *prefix, char *pattern)
{
  char cleanPrefix[4096];
  char *port_string;
  int prefixLength;
  int cleanLength;
  int i;

  char *loc_host;
  int loc_port = 0;
  char *loc_prefix = 0;
  char *loc_suffix = 0;
  int loc_is_exact = 0;
    
#ifdef WIN32
  if (host) {
    host = cse_strdup(config->p, host);
    for (i = 0; host[i]; i++)
      host[i] = tolower(host[i]);
  }
  if (prefix) {
    prefix = cse_strdup(config->p, prefix);
    for (i = 0; prefix[i]; i++)
      prefix[i] = tolower(prefix[i]);
  }
  if (pattern) {
    pattern = cse_strdup(config->p, pattern);
    for (i = 0; pattern[i]; i++)
      pattern[i] = tolower(pattern[i]);
  }
#endif /* WIN32 */

  cleanPrefix[0] = 0;

  loc_host = host ? cse_strdup(config->p, host) : 0;
  if (loc_host && (port_string = strchr(loc_host, ':'))) {
    *port_string = 0;
    loc_port = atoi(port_string + 1);
  }

  if (! prefix[0] && pattern[0] && pattern[0] != '/' && pattern[0] != '*')
    strcpy(cleanPrefix, "/");
  else if (prefix[0] && prefix[0] != '/')
    strcpy(cleanPrefix, "/");

  strcat(cleanPrefix, prefix);
  prefixLength = strlen(cleanPrefix);
  if (prefixLength > 0 && cleanPrefix[prefixLength - 1] == '/')
    cleanPrefix[prefixLength - 1] = 0;

  if (! pattern[0]) {
    loc_prefix = cse_strdup(config->p, cleanPrefix);
    loc_suffix = 0;
  }
  else if (pattern[0] == '*') {
    loc_prefix = cse_strdup(config->p, cleanPrefix);
    loc_suffix = pattern + 1;
  }
  else {
    if (pattern[0] != '/')
      strcat(cleanPrefix, "/");
    strcat(cleanPrefix, pattern);

    cleanLength = strlen(cleanPrefix);

    if (strlen(pattern) <= 1)
      cleanPrefix[cleanLength - 1] = 0;
    else if (cleanPrefix[cleanLength - 1] != '*')
      loc_is_exact = 1;
    else if (cleanLength > 2 && cleanPrefix[cleanLength - 2] == '/')
      cleanPrefix[cleanLength - 2] = 0;
    else if (cleanLength > 1)
      cleanPrefix[cleanLength - 1] = 0;

    loc_prefix = cse_strdup(config->p, cleanPrefix);
    loc_suffix = 0;
  }

  cse_add_unique_location(config, loc_host, loc_port, loc_prefix,
                          loc_suffix, loc_is_exact);
}

/**
 * Returns true if the path is an absolute path
 */
static int
is_path_absolute(char *path)
{
  return (path[0] == '/' || path[0] == '\\' ||
          (path[1] == ':' &&
           ((path[0] >= 'a' && path[0] <= 'z') ||
            (path[0] >= 'A' && path[0] <= 'Z'))));
}

static char *
cse_app_dir(config_t *config, registry_t *node, char *parent)
{
  char buf[4096];
  char *value;

  if (! node)
    return parent;
  
  value = cse_find_value(node->first, "app-dir");
  
  if (! parent || ! *parent)
    parent = ".";

  if ((! value || ! value[0]) && ! strcmp(node->key, "web-app")) {
    value = node->value;
    
    if (value && value[0] == '/')
      value++;
  }
  
  if (! value || ! value[0]) {
    LOG(("app-dir(p) %s\n", parent));
    return parent;
  }

  /* XXX: in theory only for win32 */
  if (is_path_absolute(value)) {
    LOG(("app-dir(a) %s\n", value));
    return value;
  }

  /* since this only happens on startup, the memory leak is okay */
  sprintf(buf, "%s/%s", parent, value);
  
  LOG(("app-dir(r) %s\n", buf));
  
  return cse_strdup(config->p, buf);
}

static void
cse_init_web_app_contents(config_t *config, registry_t *node,
                          char *host, char *prefix)
{
  /* This is added automatically on the Java side, but it seems
  // inappropriate to add it to the plugin side.
  // cse_add_location(config, host, "", "*.jsp");
  // cse_add_location(config, host, "", "*.xtp");

  // WEB-INF passes to Resin so Resin can hide it.
  */
  cse_add_location(config, host, prefix, "/WEB-INF/*");
  
  for (; node; node = node->next) {
    if (! strcmp(node->key, "servlet-mapping")) {
      registry_t *url_pattern = cse_next_link(node->first, "url-pattern");

      if (url_pattern && url_pattern->value)
	cse_add_location(config, host, prefix, url_pattern->value);
    }
  }

}

/**
 * Adds a configuration file as a dependency.  If the file changes,
 * the plugin will reread the configuration.
 *
 * @param config the server configuration.
 * @param path filepath to the dependent file.
 */
static void
cse_add_depend(config_t *config, char *path)
{
  struct stat st;
  depend_t *depend;
  
  depend = (depend_t *) cse_alloc(config->p, sizeof(depend_t));
  memset(depend, 0, sizeof(depend_t));
  depend->next = config->depend_list;
  config->depend_list = depend;
  depend->path = cse_strdup(config->p, path);

  if (! stat(path, &st)) {
    depend->last_modified = st.st_mtime;
    depend->last_size = st.st_size;
  }
}

/**
 * Initialize a web application.
 */
static void
cse_init_web_app(config_t *config, registry_t *node,
                 char *host, char *prefix, char *app_dir)
{
  char buf[4096];
  FILE *file;
  
  if (node)
    cse_init_web_app_contents(config, node->first, host, prefix);
  
  /* XXX: should the registry itself get updated? i.e. adding this
  // node somewhere?
  */
  sprintf(buf, "%s/WEB-INF/web.xml", app_dir);
  LOG(("web-app %s\n", buf));
  
  cse_add_depend(config, buf);
  
  file = fopen(buf, "r");
  if (file) {
    registry_t *web = cse_parse(file, config, cse_strdup(config->p, buf));
    fclose(file);

    if (web && web->first)
      cse_init_web_app_contents(config, web->first->first, host, prefix);
  }
}

/**
 * Initialize the war dir.
 */
static void
cse_init_war_dir(config_t *config, char *host, char *war_dir)
{
#ifndef WIN32
  DIR *dir;
  struct dirent *dirent;

  LOG(("war-dir %s\n", war_dir));
  
  cse_add_depend(config, war_dir);

  dir = opendir(war_dir);
  if (! dir)
    return;
  
  while ((dirent = readdir(dir))) {
    char app_dir[4096];
    char prefix[4096];
    char *name = dirent->d_name;
    struct stat st;

    sprintf(prefix, "/%s", name);
    sprintf(app_dir, "%s/%s", war_dir, name);
    
    if (! stat(app_dir, &st) && S_ISDIR(st.st_mode) && name[0] != '.') {
      LOG(("app %s\n", name));
    
      cse_init_web_app(config, 0, host, prefix, app_dir);
    }
  }
  closedir(dir);
#endif
}


/**
 * Initialize the configuration mapping for the <host> block.
 *
 * @param config configuration structure holding the parsed mapping.
 * @param node the configuration node for the <host> block.
 * @param host default host name.
 * @param app_dir current directory.
 */
static void
cse_init_host(config_t *config, registry_t *node, char *host,
              char *app_dir)
{
  char *host_dir = cse_app_dir(config, node, app_dir);
                               
  cse_init_web_app(config, node, host, "", host_dir);

  if (! node)
    return;

  for (node = node->first; node; node = node->next) {
    if (! node->value) {
    }
    else if (! strcmp(node->key, "web-app")) {
      char *prefix = node->value;
      char *subdir = cse_app_dir(config, node, host_dir);

      cse_init_web_app(config, node, host, prefix, subdir);
    }
    else if (! strcmp(node->key, "war-dir") && node->value) {
      char *subdir = node->value;
      char war_dir[8192];

      if (is_path_absolute(subdir))
        sprintf(war_dir, "%s", subdir);
      else
        sprintf(war_dir, "%s/%s", config->resin_home, subdir);

      cse_init_war_dir(config, host, war_dir);
    }
  }
}

/**
 * Initialize the dispatch mappings.
 */
static void
cse_init_server(config_t *config, registry_t *node)
{
  char *app_dir = cse_app_dir(config, node, config->resin_home);

  cse_init_host(config, node, "", config->resin_home);

  if (! node)
    return;

  for (node = node->first; node; node = node->next) {
    if (node->value && ! strcmp(node->key, "host")) {
      char *name = node->value;
      char *host_dir = cse_app_dir(config, node, app_dir);

      cse_init_host(config, node, name, host_dir);
    }
  }
}

/**
 * Adds a new backup to the configuration
 */
srun_t *
cse_add_host(config_t *config, const char *hostname, int port)
{
  return cse_add_host_int(config, hostname, port, 0, 0);
}

/**
 * Adds a new backup to the configuration
 */
srun_t *
cse_add_backup(config_t *config, const char *hostname, int port)
{
  return cse_add_host_int(config, hostname, port, 1, 0);
}

/**
 * initialize a single srun host
 */
static void
cse_init_host_node(config_t *config, registry_t *node,
		   int is_backup, int is_default)
{
  int port;
  char *host;
  registry_t *host_node = cse_next_link(node, "srun-host");
  registry_t *port_node = cse_next_link(node, "srun-port");
  srun_t *srun = 0;
  registry_t *subnode;

  if (! host_node && ! is_default)
    host_node = cse_next_link(node, "host");
  if (! port_node && ! is_default)
    port_node = cse_next_link(node, "port");
    
  if (port_node && port_node->value && *port_node->value) {
    port = atoi(port_node->value);
  }
  else
    port = 6802;
    
  if (host_node && host_node->value && *host_node->value) {
    host = host_node->value;
  }
  else
    host = "localhost";

  srun = cse_add_host_int(config, host, port, is_backup, is_default);
  if (! srun)
    return;

  subnode = cse_next_link(node, "connect-timeout");
  if (subnode && subnode->value && *subnode->value)
    srun->connect_timeout = atoi(subnode->value);
  
  subnode = cse_next_link(node, "live-time");
  if (subnode && subnode->value && *subnode->value) {
    srun->live_time = atoi(subnode->value);
    if (srun->live_time < 5)
      srun->live_time = 5;
  }
  
  subnode = cse_next_link(node, "dead-time");
  if (subnode && subnode->value && *subnode->value) {
    srun->dead_time = atoi(subnode->value);
    if (srun->dead_time < 5)
      srun->dead_time = 5;
  }
}

/**
 * Initialize all the load-balancing hosts in the configuration.
 */
static void 
cse_init_hosts(config_t *config, registry_t *top)
{
  int has_srun = 0;
  registry_t *node;
  
  for (node = cse_next_link(top, "srun");
       node;
       node = cse_next_link(node->next, "srun")) {
    has_srun = 1;
    cse_init_host_node(config, node->first, 0, 0);
  }

  for (node = cse_next_link(top, "srun-backup");
       node;
       node = cse_next_link(node->next, "srun-backup")) {
    has_srun = 1;
    cse_init_host_node(config, node->first, 1, 0);
  }

  if (! has_srun) {
    cse_init_host_node(config, top, 0, 1);
  }
}

/**
 * Logging for the configuration file.
 */
void
cse_log_config(config_t *config)
{
  location_t *loc = config->locations;
  for (; loc; loc = loc->next) {
    LOG(("cfg host:%s prefix:%s suffix:%s next:%x\n", 
         loc->host ? loc->host : "null",
         loc->prefix ? loc->prefix : "null",
         loc->suffix ? loc->suffix : "null",
         loc->next));
  }
}

static void
cse_init_error_page(config_t *config, registry_t *node)
{
  for (node = cse_next_link(node, "error-page");
       node;
       node = cse_next_link(node->next, "error-page")) {
    char *exn = cse_find_value(node->first, "exception-type");
    char *code = cse_find_value(node->first, "error-code");
    
    if (exn && ! strcmp(exn, "connection"))
      config->error_page = cse_find_value(node->first, "location");
    else if (! exn && ! code && ! config->error_page)
      config->error_page = cse_find_value(node->first, "location");
  }
}

/**
 * Initialize the configuration.
 */
void
cse_init_config(config_t *config)
{
  registry_t *root = 0;
  registry_t *node = 0;
  FILE *is;
  char *caucho_status;
  int i;

  LOG(("initializing from %s\n", config->path));

  if (! config->p)
    config->p = cse_create_pool();

  if (! config->resin_home) {
    int p = strlen(config->path);
    
    for (;
         p >= 0 && (config->path[p] != '/' && config->path[p] != '\\');
         p--) {
    }
    
    for (p--;
         p >= 0 && (config->path[p] != '/' && config->path[p] != '\\');
         p--) {
    }

    if (p >= 0) {
      config->resin_home = strdup(config->path);
      config->resin_home[p] = 0;
    }
    else {
      config->resin_home = ".";
    }
  }

  /*
  // XXX: in theory, should free these, but they only change during
  // development, so it really shouldn't matter
  */
  config->registry = 0;
  config->locations = 0;

  for (i = 0; i < CACHE_SIZE; i++) {
    if (g_url_cache[i].uri) {
      free(g_url_cache[i].uri);
      free(g_url_cache[i].host);
    }
    g_url_cache[i].uri = 0;
  }
  
  config->update_count++;

  config->disable_caucho_status = 0;
  config->srun_list = 0;
  config->srun_capacity = 0;
  config->srun_size = 0;
  config->depend_list = 0;
  cse_add_depend(config, config->path);

  is = fopen(config->path, "r");
  if (! is) {
    cse_error(config, "Resin cannot open config file `%s'\n", config->path);
    return;
  }

  root = cse_parse(is, config, config->path);
  fclose(is);

  config->registry = root;
  cse_print_registry(root);
  
  if (root)
    node = cse_next_link(root->first, "caucho.com");
  if (node)
    node = cse_next_link(node->first, "http-server");

  if (node) {
    caucho_status = cse_find_value(node->first, "caucho-status");
	if (caucho_status &&
            (! strcmp(caucho_status, "false") ||
             ! strcmp(caucho_status, "no")))
          config->disable_caucho_status = 1;
  }

  if (node)
    cse_init_server(config, node);

  if (node)
    cse_init_hosts(config, node->first);

  if (node)
    cse_init_error_page(config, node->first);

  if (! config->lock)
    config->lock = cse_create_lock();
}

/**
 * Check the configuration files for any changes.  If there are any changes,
 * reload the configuration files.
 */
void
cse_update_config(config_t *config, int now)
{
  struct stat st;
  depend_t *depend;
  int has_changed = 0;
  int diff = config->last_update - now;

  if (diff > -15 && diff < 15)
    return;

  config->last_update = now;
  
  for (depend = config->depend_list;
       depend && ! has_changed;
       depend = depend->next) {
    if (stat(depend->path, &st))
      has_changed = depend->last_modified != 0;
    else if (depend->last_modified != st.st_mtime)
      has_changed = 1;
    else if (depend->last_size != st.st_size)
      has_changed = 1;

    if (has_changed) {
      LOG(("%s mod:%d->%d size:%d->%d\n", depend->path,
           depend->last_modified, st.st_mtime,
           depend->last_size, st.st_size));
    }
  }

  if (has_changed) {
    cse_lock(config->lock);
    cse_close_sockets(config);
    cse_free_pool(config->p);
    cse_init_config(config);
    cse_unlock(config->lock);
  }
}


/**
 * tests if 'full' starts with 'part'
 *
 * If it's not an exact match, the next character of 'full' must be '/'.
 * That way, a servlet mapping of '/foo/ *' will match /foo, /foo/bar, but
 * not /foolish.
 */
static int
cse_starts_with(const char *full, const char *part)
{
  char ch1, ch2;

  while ((ch2 = *part++) && (ch1 = *full++) && ch1 == ch2) {
  }

  if (ch2)
    return 0;
  else if (! *full)
    return 1;
  else
    return *full == '/';
}

static int
cse_match_suffix(const char *full, const char *suffix)
{
  int len = strlen(suffix);

  do {
    char *match = strstr(full, suffix);
    if (! match)
      return 0;
    
    if (! match[len] || match[len] == '/')
      return 1;
#ifdef WIN32
    if (match[len] == '.') {
      for (; match[len] == '.'; len++) {
      }

      if (! match[len] || match[len] == '/')
     	return 1;
    }
#endif

    full = match + len;
  } while (*full);

  return 0;
}

/**
 * Match the url to the location pattern.
 *
 * The match is cached by url_cache so common urls will be served quickly.
 *
 * @param loc the location pattern.
 * @param host the request's host
 * @param port the request's port
 * @param uri the request's uri
 */

static int
cse_match_location(config_t *config,
                   location_t *loc,
                   const char *host, int port,
                   const char *uri)
{
  int hash = port;
  int i;
  int is_match = 0;
  
  hash_t *entry;
  char *test_uri;
  char *test_host;
  int test_port;
  int test_count;
  int test_match;

  if (! host)
    host = "";
  if (! uri)
    uri = "";
  
  for (i = 0; host && host[i]; i++)
    hash = 65531 * hash + host[i];
  
  for (i = 0; uri && uri[i]; i++)
    hash = 65531 * hash + uri[i];

  if (hash < 0)
    hash = -hash;
  hash = hash % CACHE_SIZE;

  entry = &g_url_cache[hash];

  test_count = entry->count;
  test_uri = entry->uri;
  test_host = entry->host;
  test_port = entry->port;
  test_match = entry->is_match;

  if (test_count != entry->count) {
  }
  else if (! test_uri || strcmp(test_uri, uri)) {
  }
  else if (! test_host || strcmp(test_host, host)) {
  }
  else if (test_port && test_port != port) {
  }
  else
    return test_match;
  
  for (; loc; loc = loc->next) {
    LOG(("match host:%s:%d prefix:%s suffix:%s with host:%s uri:%s next:%x\n", 
         loc->host ? loc->host : "null",
         loc->port,
         loc->prefix ? loc->prefix : "null",
         loc->suffix ? loc->suffix : "null",
         host, uri, loc->next));
    
    if (*(loc->host) && ! strstr(loc->host, host))
      continue;

    else if (port && loc->port && port != loc->port)
      continue;

    else if (loc->is_exact && ! strcmp(uri, loc->prefix)) {
      is_match = 1;
      break;
    }
    else if (loc->is_exact)
      continue;
      
    else if (! cse_starts_with(uri, loc->prefix))
      continue;
    
    else if (loc->suffix && ! cse_match_suffix(uri, loc->suffix))
      continue;

    is_match = 1;
    break;
  }

  if (! is_match && strstr(uri, "/j_security_check"))
    is_match = 1;

  cse_lock(config->lock);
  entry->count++;
  entry->is_match = is_match;
  if (entry->uri) {
    free(entry->host);
    free(entry->uri);
    
    entry->uri = 0;
    entry->host = 0;
  }

  LOG(("entry %s %s\n", host, uri));
  
  entry->host = strdup(host);
  entry->uri = strdup(uri);
  entry->port = port;
  entry->count++;
  cse_unlock(config->lock);

  return is_match;
}

/**
 * Tests if the request matches a Resin URL.
 *
 * @param config the plugin configuration structure
 * @param host the request's host
 * @param port the request's port
 * @param url the request's uri
 */
int
cse_match_request(config_t *config, const char *host,
                  int port, const char *uri)
{
  char buf[16 * 1024];
  int len = sizeof(buf) - 1;
  int i, k;
  int ch;

  k = 0;
  for (i = 0; (ch = uri[i]) && i + 1 < len; i++) {
    if ((ch == ';' || ch == '?') &&
        ! strncmp(uri + i + 1, "jsessionid=", sizeof("jsessionid=") - 1)) {
      break;
    }
    else if (ch == '/' && k > 0 && buf[k - 1] == '/')
      continue;
    else {
#ifdef WIN32
      buf[k++] = tolower(ch);
#else
      buf[k++] = ch;
#endif
    }
  }
  buf[k] = 0;

  return (cse_match_location(config, config->locations, host, port, buf));
}
