/*
 * Copyright (c) 1999-2000 Caucho Technology.  All rights reserved.
 *
 * Caucho Technology permits redistribution, modification and use
 * of this file in source and binary form ("the Software") under the
 * Caucho Developer Source License ("the License").  The following
 * conditions must be met:
 *
 * 1. Each copy or derived work of the Software must preserve the copyright
 *    notice and this notice unmodified.
 *
 * 2. Redistributions of the Software in source or binary form must include 
 *    an unmodified copy of the License, normally in a plain ASCII text
 *
 * 3. The names "Resin" or "Caucho" are trademarks of Caucho Technology and
 *    may not be used to endorse products derived from this software.
 *    "Resin" or "Caucho" may not appear in the names of products derived
 *    from this software.
 *
 * 4. Caucho Technology requests that attribution be given to Resin
 *    in any manner possible.  We suggest using the "Resin Powered"
 *    button or creating a "powered by Resin(tm)" link to
 *    http://www.caucho.com for each page served by Resin.
 *
 * This Software is provided "AS IS," without a warranty of any kind. 
 * ALL EXPRESS OR IMPLIED REPRESENTATIONS AND WARRANTIES, INCLUDING ANY
 * IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE
 * OR NON-INFRINGEMENT, ARE HEREBY EXCLUDED.

 * CAUCHO TECHNOLOGY AND ITS LICENSORS SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE OR ANY THIRD PARTY AS A RESULT OF USING OR
 * DISTRIBUTING SOFTWARE. IN NO EVENT WILL Caucho OR ITS LICENSORS BE LIABLE
 * FOR ANY LOST REVENUE, PROFIT OR DATA, OR FOR DIRECT, INDIRECT, SPECIAL,
 * CONSEQUENTIAL, INCIDENTAL OR PUNITIVE DAMAGES, HOWEVER CAUSED AND
 * REGARDLESS OF THE THEORY OF LIABILITY, ARISING OUT OF THE USE OF OR
 * INABILITY TO USE SOFTWARE, EVEN IF HE HAS BEEN ADVISED OF THE POSSIBILITY
 * OF SUCH DAMAGES.      
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <netinet/in.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <dirent.h>
#include <time.h>
#include <signal.h>
/* probably system-dependent */
#include <dlfcn.h>
#include <jni.h>

#include "cse.h"
#include "proxy.h"
#include "proxy_jni.h"

#define RING_SIZE 32

typedef struct t_java_req {
  jobject request_obj;
  jbyteArray uri;
  jcharArray header_buffer;
  jintArray header_offsets;
} t_java_req;

struct server_t {
  int ss;

  JNIEnv *env;
  jobject server_obj;

  int is_dead;

  pthread_t accept_thread;

  pthread_mutex_t lock;
  pthread_cond_t accept_cond;

  /* Socket ring buffer from the accept thread to the connections */
  int accept_sockets[RING_SIZE];
  /* Address ring buffer from the accept thread to the connections */
  struct sockaddr_in accept_addresses[RING_SIZE];
  /* Head of the ring buffer */
  int accept_head;
  /* Tail of the ring buffer */
  int accept_tail;

  /* Maximum number of connections allowed waiting for an accept */
  int accept_max;
  /* Current number of connections waiting for an accept */
  int accept_count;

  /* Maximum number of connections */
  int connection_max;
  /* Current number of connections */
  int connection_count;

  /* Debug id */
  int connection_id;
  
  connection_t *conn_head;
};

int g_now_seconds;
server_t *g_server;

static void
fill_fields(JNIEnv *env, jobject server_obj, t_java_req *java_req)
{
  jobject request_obj;
  
  request_obj = (*env)->CallObjectMethod(env, server_obj,
                                         g_java.create_request);
  if ((*env)->ExceptionOccurred(env)) {
    (*env)->ExceptionDescribe(env);
    exit(1);
  }
  request_obj = (*env)->NewGlobalRef(env, request_obj);
  
  java_req->request_obj = request_obj;

  java_req->uri = (*env)->GetObjectField(env, request_obj, g_java.f_uri);
  
  java_req->header_buffer = (*env)->GetObjectField(env, request_obj,
                                                   g_java.f_header_buffer);
  java_req->header_offsets = (*env)->GetObjectField(env, request_obj,
                                                    g_java.f_header_offsets);
}

/**
 * Returns true if the server will allow a HTTP/1.1 keepalive
 */
int
server_allow_keepalive(server_t *server)
{
  return (server->connection_count < server->connection_max - 50 ||
          server->accept_count >= server->accept_max);
}

/**
 * Handle a connection from a user.
 *
 * @param sock the opened socket.
 * @param req the request object
 * @param res the response object
 * @param env the Java thread environment
 * @param java_req pointers to the java request.
 */
static void
do_request(int sock, request_t *req, response_t *res,
           JNIEnv *env, t_java_req *java_req)
{
  int len;
  char buf[4096];
  char *uri;
  char *ptr;
  short *headers;
  char *header_ptr;
  int *header_offsets;
  int header_start;
  jobject request_obj = java_req->request_obj;
  jint local_addr;
  jint local_port;
  jint remote_addr;
  jint remote_port;
    
  int i, j, k;

  local_addr = ntohl(req->server_addr.sin_addr.s_addr);
  local_port = ntohs(req->server_addr.sin_port);
  remote_addr = ntohl(req->remote_addr.sin_addr.s_addr);
  remote_port = ntohs(req->remote_addr.sin_port);

  do {
    response_start(res, sock);
    if (request_parse(req, sock) < 0)
      break;

    uri = (*env)->GetPrimitiveArrayCritical(env, java_req->uri, 0);
    ptr = req->uri;
    for (len = 0; ptr[len]; len++)
      uri[len] = ptr[len];

    (*env)->ReleasePrimitiveArrayCritical(env, java_req->uri, uri, 0);
    (*env)->SetIntField(env, request_obj, g_java.f_uri_length, len);

    headers = (*env)->GetPrimitiveArrayCritical(env, java_req->header_buffer, 0);
    header_offsets = (*env)->GetPrimitiveArrayCritical(env,
                                                       java_req->header_offsets,
                                                       0);

    i = 0;
    j = 0;
  
    header_offsets[j++] = 0;
    for (k = 0; req->method[k]; k++)
      headers[i++] = req->method[k];
  
    header_offsets[j++] = i;
    for (k = 0; req->protocol[k]; k++)
      headers[i++] = req->protocol[k];

    header_offsets[j++] = i;
    for (k = 0; req->host && req->host[k]; k++)
      headers[i++] = req->host[k];

    header_offsets[j++] = i;
    /* null for later */

    header_start = i;
  
    len = req->header_length;
    header_ptr = req->headers;
    for (k = 0; k < len; k++)
      headers[i++] = header_ptr[k];
  
    len = req->header_size;
    for (k = 0; k <= len; k++) {
      header_offsets[j++] = header_start + req->header_keys[k];
      header_offsets[j++] = header_start + req->header_values[k];
    }
  
    (*env)->ReleasePrimitiveArrayCritical(env, java_req->header_buffer,
                                          headers, 0);
  
    (*env)->ReleasePrimitiveArrayCritical(env, java_req->header_offsets,
                                          header_offsets, 0);
    (*env)->SetIntField(env, request_obj, g_java.f_header_size, len);

    (*env)->CallVoidMethod(env, request_obj, g_java.dispatch,
                           req, local_addr, local_port,
                           remote_addr, remote_port);

    if ((*env)->ExceptionOccurred(env)) {
      (*env)->ExceptionDescribe(env);
    }

    cse_flush_request(res);
  } while (req->keepalive);
  
  close(sock);
}

/**
 * Creates a new connection.  If there are already too many connections,
 * return null.
 */
static connection_t *
connection_create(server_t *server)
{
  connection_t *conn;
  int id;
  int count;


  pthread_mutex_lock(&server->lock);
  if (server->connection_count > server->connection_max) {
    pthread_mutex_unlock(&server->lock);
    return 0;
  }
  count = server->connection_count++;
  id = server->connection_id++;
  pthread_mutex_unlock(&server->lock);

  /*
  fprintf(stderr, "alloc conn id=%d count=%d\n", id, count);
  */
  
  conn = (connection_t *) cse_malloc(sizeof(connection_t));
  memset(conn, 0, sizeof(connection_t));
  
  conn->id = id;
  conn->server = server;
  conn->request = request_create();
  conn->response = response_create();
  conn->request->server = server;

  conn->request->response = conn->response;
  conn->response->request = conn->request;

  pthread_mutex_lock(&server->lock);
  
  conn->next = server->conn_head;
  if (server->conn_head)
    server->conn_head->prev = conn;
  server->conn_head = conn;
  
  pthread_mutex_unlock(&server->lock);

  DEBUG(("create id=%d count=%d\n", id, count));

  return conn;
}

/**
 * Frees the connection.
 *
 * @param conn the connection to be freed.
 */
static void
connection_free(connection_t *conn)
{
  int count;
  connection_t *next;
  connection_t *prev;
  
  pthread_mutex_lock(&conn->server->lock);
  count = conn->server->connection_count--;

  next = conn->next;
  prev = conn->prev;

  if (next)
    next->prev = prev;
  if (prev)
    prev->next = next;
  else
    conn->server->conn_head = next;
  
  pthread_mutex_unlock(&conn->server->lock);

  DEBUG(("free id=%d count=%d\n", conn->id, count));

  request_free(conn->request);
  response_free(conn->response);
  cse_free(conn);
}

/**
 * Main thread loop for the request
 *
 * @param v_server pointer to the server object
 */
static void *
request_run(void *v_connection)
{
  connection_t *conn = v_connection;
  server_t *server = conn->server;
  request_t *req = conn->request;
  response_t *res = conn->response;
  t_java_req java_req;
  JNIEnv *env;
  int result;

  // signal(SIGALRM, alarm_handler);

  DEBUG((stderr, "new connection\n"));

  result = (*g_java.vm)->AttachCurrentThread(g_java.vm, (void**)&env, 0);

  fill_fields(env, server->server_obj, &java_req);
  
  while (! server->is_dead && server->accept_count <= server->accept_max) {
    struct sockaddr_in *sin;
    int sock = -1;
    int tail;

    pthread_mutex_lock(&server->lock);
    tail = server->accept_tail;

    if (server->accept_head == tail) {
      server->accept_count++;
      pthread_cond_wait(&server->accept_cond, &server->lock);
      server->accept_count--;

      tail = server->accept_tail;
    }
      
    if (server->accept_head != tail) {
      sock = server->accept_sockets[tail];
      req->remote_addr = server->accept_addresses[tail];
      server->accept_tail = (tail + 1) % RING_SIZE;
    }
    pthread_mutex_unlock(&server->lock);

    if (sock >= 0) {
      int len = sizeof(req->server_addr);
      getsockname(sock, (struct sockaddr *) &req->server_addr, &len);
    
      do_request(sock, req, res, env, &java_req);
    }
  }

  if (java_req.request_obj)
    (*env)->DeleteGlobalRef(env, java_req.request_obj);

  (*g_java.vm)->DetachCurrentThread(g_java.vm);

  connection_free(conn);

  return 0;
}

static void
start_connection(server_t *server)
{
  connection_t *conn = connection_create(server);
  int result;

  if (conn) {
    result = pthread_create(&conn->thread, 0, request_run, conn);
    
    if (result < 0) {
      pthread_mutex_lock(&server->lock);
      server->connection_count--;
      pthread_mutex_unlock(&server->lock);
    }
  }
}

/**
 * Main loop for the accept server.
 *
 * @param server the server variable
 */
static void *
accept_run(void *v_server)
{
  server_t *server = v_server;
  while (! server->is_dead) {
    int head =  server->accept_head;
    int next_head = (head + 1) % RING_SIZE;
    struct sockaddr_in sin;
    int len = sizeof(sin);
    int sock;
    int needs_new_connection = 0;

    memset(&sin, 0, sizeof(sin));
    sock = accept(server->ss, (struct sockaddr *) &sin, &len);

    // If overflow, kill it quickly
    if (next_head == server->accept_tail) {
      close(sock);
      continue;
    }

    server->accept_sockets[head] = sock;
    server->accept_addresses[head] = sin;
    server->accept_head = next_head;
    pthread_mutex_lock(&server->lock);
    if (server->accept_count > 0) {
      pthread_cond_signal(&server->accept_cond);
    }
    else {
      needs_new_connection = 1;
    }
    pthread_mutex_unlock(&server->lock);

    if (needs_new_connection)
      start_connection(server);
  }

  return 0;
}

/**
 * Starts the server
 *
 * @param ss the server socket file descriptor
 * @param env the JDK environment
 * @param server_obj the Java JniServer object
 *
 * @return the C-based server object.
 */
server_t *
start_server(int ss, JNIEnv *env, jobject server_obj)
{
  server_t *server = cse_malloc(sizeof(server_t));
  int i;
  int result;
  pthread_t alarm_thread;
  
  memset(server, 0, sizeof(server_t));
  server->ss = ss;
  server->env = env;
  server->accept_max = 5;
  server->connection_max = 100;
  server->server_obj = server_obj;

  pthread_mutex_init(&server->lock, 0);
  pthread_cond_init(&server->accept_cond, 0);

  g_now_seconds = time(0);
    
  result = pthread_create(&server->accept_thread, 0, accept_run, server);

  return server;
}

int
server_is_dead(server_t *server)
{
  return ! server || server->is_dead;
}

void
close_server(server_t *server)
{
  fprintf(stderr, "closing server\n");
  
  server->is_dead = 1;

  pthread_mutex_lock(&server->lock);
  pthread_cond_broadcast(&server->accept_cond);
  pthread_mutex_unlock(&server->lock);
}
